require 'store'

# Models #######################################################################
class User
  attr_accessor :name
end

class Catalog 
  attr_accessor :editor, :name, :products

  def products
    @products ||= []
  end

  def add_product(product)
    products << product
  end
end

class Product
  attr_accessor :name, :price, :discount

  def discounted_price
    price * (1 - discount)
  end

  def discount
    @discount || 0
  end
end

# EntityMappers ################################################################

class UserEntityMapper < Store::EntityMapper
  def map(user)
    {
      'name' => user.name
    }
  end

  def unmap(cls, mapped)
    user = cls.new
    user.name = mapped['name']
    user
  end
end

class CatalogEntityMapper < Store::EntityMapper
  def map(catalog)
    {
      'editor' => store.save(catalog.editor),
      'name' => catalog.name,
      'products' => store.save(catalog.products)
    }
  end

  def unmap(cls, mapped)
    cls.new.tap do |entity|
      entity.editor   = store.load(mapped['editor'])
      entity.name     = mapped['name']
      entity.products = store.load(mapped['products'])
    end
  end
end

class ProductEntityMapper < Store::EntityMapper
  def map(product)
    {
      'name' => product.name,
      'price' => product.price,
      'discount' => product.discount
    }
  end

  def unmap(cls, mapped)
    product = cls.new
    product.name = mapped['name']
    product.price = mapped['price']
    product.discount = mapped['discount']
    product
  end
end

# DataMappers###################################################################

class MemoryDataMapper < Store::DataMapper
  def initialize
    @items = {}
    @id = 0
  end

  def insert(data)
    id = next_id
    @items[id] = dup(data)
    id
  end

  def update(id, data)
    @items[id] = dup(data)
    id
  end

  def delete(id)
    @items.delete(id)
    id
  end

  def single_find(id)
    dup(@items[id])
  end

  def bulk_find(ids)
    @items.values_at(*ids).map do |data|
      dup data
    end
  end

  private

  def all
    @items
  end

  def next_id
    @id += 1
  end

  def dup(data)
    data and Marshal.load(Marshal.dump(data))
  end
end

class UserDataMapper < MemoryDataMapper
end

class CatalogDataMapper < MemoryDataMapper

  def select(query)
    if query.name == :find_by_name
      select_by_name(query.args.first)
    else
      raise "Query not known"
    end
  end

  private

  def select_by_name(name)
    all.select do |id, data|
      data['name'] == name
    end
  end
end

class ProductDataMapper < MemoryDataMapper
end

# Store instantiation ##########################################################

entities = {
  'User'    => User,
  'Catalog' => Catalog,
  'Product' => Product
}

data_mappers = {
  'User'    => UserDataMapper.new,
  'Catalog' => CatalogDataMapper.new,
  'Product' => ProductDataMapper.new
}

entity_mappers = {
  'User'    => UserEntityMapper,
  'Catalog' => CatalogEntityMapper,
  'Product' => ProductEntityMapper
}

class Repo < Store::Repo
  def find_catalogs_by_name(name)
    query Store::Query.new('Catalog', :find_by_name, name)
  end
end

class DiscountedProductFactory < Struct.new(:repo)
  def build(name, price = 100)
    repo.build 'Product', :name => name, :price => price, :discount => 0.25
  end
end

repo = Repo.new(entities, data_mappers, entity_mappers)

# Usage ########################################################################

editor = repo.build 'User', :name => 'Mr. Repo'

catalog = repo.build 'Catalog', :name => 'Catalog 1', :editor => editor

factory = DiscountedProductFactory.new(repo)

catalog.add_product repo.build('Product', :name => 'Pickaxe', :price => 100)
catalog.add_product repo.build('Product', :name => 'Scredriver', :price => 400)
catalog.add_product factory.build('Pair of shoes')

p 'save call'
repo.save(catalog)

catalog.products[0].price = 24

p 'save call'
repo.save(catalog)

product = catalog.products.delete_at(1)
repo.remove product

p catalog

p 'save call'
# no implicit delete here. is this a problem?
repo.save(catalog)

p repo.find_catalogs_by_name('test')

p repo.find_catalogs_by_name('Catalog 1')

# simulate external change

items = data_mappers['Catalog'].instance_variable_get(:@items)
items[1]['name'] = 'Externally updated Catalog'

# reload known entity
p repo.reload(catalog)
