require 'store/repo'
require 'store/ref'
require 'store/entity_mapper'
require 'store/data_mapper'
require 'store/resolver'
require 'store/entity_references'

class Store
  def initialize(entity_classes, data_mapping, entity_mapping)
    @entity_classes         = entity_classes
    @resolver_data_mapper   = Resolver::DataMapper.new(data_mapping)
    @resolver_entity_mapper = Resolver::EntityMapper.new(self, entity_mapping)
    @entity_references      = EntityReferences.new
  end

  def build(entity_class_name, attributes = {})
    entity = @entity_classes[entity_class_name].new

    attributes.each do |attribute, value|
      entity.public_send :"#{attribute}=", value
    end

    entity
  end

  def save(entities)
    if entities.kind_of? Array
      entities.map do |entity|
        save_entity(entity)
      end
    else
      entity = entities
      save_entity(entity)
    end
  end

  def remove(entity)
    data_mapper = data_mapper(entity)
    raise_unless_data_mapper_found(entity, data_mapper)

    reference = find_entity_reference(entity)
    data_mapper.delete(reference)

    remove_entity_reference entity
  end

  def query(query)
    data_mapper = data_mapper(query)
    entity_class = @entity_classes[query.entity]

    raise_unless_data_mapper_found(query, data_mapper)

    mapped_entities_with_references = data_mapper.select(query) || {}

    mapped_entities_with_references.map do |reference, mapped_entity|
      entity = entity_mapper(query).unmap(entity_class, mapped_entity)
      save_entity_reference reference, entity
      entity
    end
  end

  def load(refs)
    if refs.kind_of? Array
      bulk_load refs
    else
      single_load refs
    end
  end

  def reload(entity)
    reference = find_entity_reference(entity)
    raise "Unknown entity #{entity.inspect}" unless reference
    ref = Store::Ref.new(extract_entity_class_name(entity), reference)
    single_load(ref)
  end
  
  def single_load(ref)
    data_mapper = data_mapper(ref)
    entity_class = @entity_classes[ref.entity_type]
    reference = ref.reference

    mapped_entity = data_mapper.single_find(reference)
    raise "Entity not found for reference #{ref.inspect}" unless mapped_entity

    entity = entity_mapper(ref).unmap(entity_class, mapped_entity)
    save_entity_reference reference, entity
    entity
  end

  def bulk_load(refs)
    data_mapper = data_mapper(refs.first)
    entity_class = @entity_classes[refs.first.entity_type]

    references = refs.map(&:reference)
    mapped_entities = data_mapper.bulk_find(references)

    entities = []

    references.each_with_index do |reference, i|
      mapped_entity = mapped_entities[i]

      entity = entity_mapper(refs.first).unmap(entity_class, mapped_entity)
      save_entity_reference reference, entity
      entities << entity
    end

    entities
  end

  def reference(entity)
    reference = find_entity_reference(entity)
    raise "Unknown entity #{entity.inspect}" unless reference
    build_reference entity, reference if reference
  end

  private
  def save_entity(entity)
    return nil unless entity

    data_mapper = data_mapper(entity)
    raise_unless_data_mapper_found(entity, data_mapper)

    reference = find_entity_reference(entity)
    if reference.nil?
      reference = data_mapper.insert(extract_entity_attributes(entity))
      save_entity_reference(reference, entity)
    else
      data_mapper.update(reference, extract_entity_attributes(entity))
    end

    build_reference entity, reference
  end
  
  def build_reference(entity, reference)
    Ref.new(extract_entity_class_name(entity), reference)
  end

  def raise_unless_data_mapper_found(source, data_mapper)
    raise "No data_mapper for #{source.inspect} found" unless data_mapper
  end

  def data_mapper(obj)
    @resolver_data_mapper.resolve_data_mapper(obj)
  end

  def entity_mapper(obj)
    @resolver_entity_mapper.resolve_entity_mapper(obj)
  end

  def extract_entity_attributes(entity)
    entity_mapper(entity).map(entity)
  end

  def extract_entity_class_name(obj)
    @resolver_entity_mapper.extract_entity_class_name_from_object(obj)
  end

  def find_entity_reference(entity)
    @entity_references.find(entity)
  end

  def save_entity_reference(reference, entity)
    @entity_references.save(reference, entity)
  end

  def remove_entity_reference(entity)
    @entity_references.remove entity
  end

end
